@@ -12,49 +12,49 @@
 
 void reflex_code_FSM(reflex::Matcher& m)
 {
-  int c0 = 0, c1 = 0;
-  m.FSM_INIT(c1);
+  int c = 0;
+  m.FSM_INIT(c);
 
 S0:
   m.FSM_FIND();
-  c1 = m.FSM_CHAR();
-  if (c1 == 'e') goto S2;
-  return m.FSM_HALT(c1);
+  c = m.FSM_CHAR();
+  if (c == 'e') goto S2;
+  return m.FSM_HALT(c);
 
 S2:
-  c1 = m.FSM_CHAR();
-  if (c1 == 'l') goto S4;
-  return m.FSM_HALT(c1);
+  c = m.FSM_CHAR();
+  if (c == 'l') goto S4;
+  return m.FSM_HALT(c);
 
 S4:
-  c1 = m.FSM_CHAR();
-  if (c1 == 'i') goto S7;
-  if (c1 == 'e') goto S9;
-  return m.FSM_HALT(c1);
+  c = m.FSM_CHAR();
+  if (c == 'i') goto S7;
+  if (c == 'e') goto S9;
+  return m.FSM_HALT(c);
 
 S7:
-  c1 = m.FSM_CHAR();
-  if (c1 == 't') goto S11;
-  return m.FSM_HALT(c1);
+  c = m.FSM_CHAR();
+  if (c == 't') goto S11;
+  return m.FSM_HALT(c);
 
 S9:
-  c1 = m.FSM_CHAR();
-  if (c1 == 'i') goto S13;
-  return m.FSM_HALT(c1);
+  c = m.FSM_CHAR();
+  if (c == 'i') goto S13;
+  return m.FSM_HALT(c);
 
 S11:
-  c0 = c1, c1 = m.FSM_CHAR();
-  if (m.FSM_META_EWE(c0, c1)) {
+  c = m.FSM_CHAR();
+  if (m.FSM_META_EWE(c)) {
     if (m.FSM_META_BWB()) {
-      m.FSM_TAKE(1, c1);
+      m.FSM_TAKE(1, c);
     }
   }
-  return m.FSM_HALT(c1);
+  return m.FSM_HALT(c);
 
 S13:
-  c1 = m.FSM_CHAR();
-  if (c1 == 'f') goto S17;
-  return m.FSM_HALT(c1);
+  c = m.FSM_CHAR();
+  if (c == 'f') goto S17;
+  return m.FSM_HALT(c);
 
 S15:
   if (m.FSM_META_BWB()) {
@@ -63,27 +63,27 @@ S15:
   return m.FSM_HALT();
 
 S17:
-  c1 = m.FSM_CHAR();
-  if (c1 == 'e') goto S21;
-  return m.FSM_HALT(c1);
+  c = m.FSM_CHAR();
+  if (c == 'e') goto S21;
+  return m.FSM_HALT(c);
 
 S19:
   m.FSM_TAKE(1);
   return m.FSM_HALT();
 
 S21:
-  c1 = m.FSM_CHAR();
-  if (c1 == 'n') goto S23;
-  return m.FSM_HALT(c1);
+  c = m.FSM_CHAR();
+  if (c == 'n') goto S23;
+  return m.FSM_HALT(c);
 
 S23:
-  c1 = m.FSM_CHAR();
-  if (c1 == 'd') goto S11;
-  return m.FSM_HALT(c1);
+  c = m.FSM_CHAR();
+  if (c == 'd') goto S11;
+  return m.FSM_HALT(c);
 }
 
 extern const reflex::Pattern::Pred reflex_pred_FSM[4100] = {
-    2,  2,101,108,
+    2,130,101,108,
     0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
